//: Playground - noun: a place where people can play

import Foundation
import Combine
import PlaygroundSupport
PlaygroundPage.current.needsIndefiniteExecution = true

/* ------------- START: Object definitions ------------- */

/// Representation of a timer that can be used to display countdown
struct Countdown {
    let seconds: Int
    let minutes: Int
    let hours: Int
    let days: Int
}

extension Countdown {
    /// Create a timer from time interval
    ///
    /// - Parameter interval: time interval in seconds
    /// - Returns: new Timer object
    static func from(interval: TimeInterval) -> Countdown {
        let seconds = Int(interval.truncatingRemainder(dividingBy: 60))
        let minutes = Int((interval / 60).truncatingRemainder(dividingBy: 60))
        let hours = Int((interval / 3600).truncatingRemainder(dividingBy: 24))
        let days = Int((interval / (24 * 3600)))
        return Countdown(seconds: seconds, minutes: minutes, hours: hours, days: days)
    }

    /// Print time for debuging
    func printTime() {
        print("\(self.days) days, \(self.hours) hours, \(self.minutes) minutes, \(self.seconds) seconds")
    }
}
/* ------------- END: Object definitions ------------- */

var cancellables = Set<AnyCancellable>()

/// End date as it would be provided by a remote database
let endDate = CurrentValueSubject<Date, Never>(Date.distantFuture)
/// The observable provided by the "database"
let endDateObsevable = endDate
    .handleEvents(receiveOutput: { print("New endDate set \($0)") })
    .eraseToAnyPublisher()

/// How often do we want to update the timer view?
///
/// For demo 0.5 is fine. For displaying this on end user device 0.1 or 0.05 would be better
/// to ensure we don't skip a second as that would look bad
let SAMPLING_RATE = 0.5

/// Sampling observable that generates an event based on SAMPLING_RATE
let sampling = Timer.TimerPublisher(interval: SAMPLING_RATE, runLoop: .main, mode: .default).autoconnect()

/// Combine Magic happening here
//endDateObsevable
//    .removeDuplicates()
//    .map { (endDate) in
//        sampling.map { _ in
//            endDate
//        }
//    }
//    .switchToLatest()
//    .map { (endDate) in
//        endDate.timeIntervalSinceNow
//    }
//    .map { (interval) -> Countdown in
//        Countdown.from(interval: interval)
//    }
//    .sink { (timer) in
//        timer.printTime()
//    }
//    .store(in: &cancellables)

sampling.combineLatest(endDateObsevable.removeDuplicates())
    .map { (_, endDate) in
        return endDate.timeIntervalSinceNow
    }
    .map { (interval) -> Countdown in
        return Countdown.from(interval: interval)
    }
    .sink { (timer) in
        timer.printTime()
    }
    .store(in: &cancellables)


/// Simulate editing of the end dates in a remote database
DispatchQueue.main.asyncAfter(deadline: DispatchTime.now() + 10) {
    endDate.send(Date().addingTimeInterval(TimeInterval(100)))
}

DispatchQueue.main.asyncAfter(deadline: DispatchTime.now() + 60) {
    endDate.send(Date().addingTimeInterval(TimeInterval(100)))
}
